from collections import Counter

import numpy as np
import pytest
import xarray as xr

from xwrf.destagger import _destag_variable, _drop_attrs, _rename_staggered_coordinate


@pytest.mark.parametrize(
    'input_attrs, output_attrs',
    [
        ({'a': 0, 'b': 1, 'c': 2}, {'a': 0, 'c': 2}),
        ({'b': 3}, {}),
        ({'a': 4, 'c': 5}, {'a': 4, 'c': 5}),
        (Counter(('a', 'a', 'b', 'c')), {'a': 2, 'c': 1}),
    ],
)
def test_drop_attrs_successful(input_attrs, output_attrs):
    result = _drop_attrs(input_attrs, ('b',))
    assert result == output_attrs
    assert isinstance(result, dict)


def test_drop_attrs_unsuccessful():
    assert _drop_attrs('not a Mapping', ('a',)) is None


@pytest.mark.parametrize(
    'input_name, stagger_dim, unstag_dim_name, output_name',
    [
        (
            'bottom_top',
            'bottom_top_stag',
            None,
            'bottom_top',
        ),
        ('bottom_top_stag', 'bottom_top_stag', 'z', 'z'),
        ('bottom_top_stag', 'bottom_top_stag', None, 'bottom_top'),
        ('XLAT_U', 'west_east', None, 'XLAT'),
        ('XLONG_V', 'south_north', None, 'XLONG'),
    ],
)
def test_rename_staggered_coordinate(input_name, stagger_dim, unstag_dim_name, output_name):
    assert _rename_staggered_coordinate(input_name, stagger_dim, unstag_dim_name) == output_name


def test_destag_variable_dataarray():
    with pytest.raises(ValueError) as exc_info:
        mock_da = xr.DataArray(
            np.zeros((2, 2)), dims=('x_stag', 'y'), coords={'x_stag': [0, 1], 'y': [0, 1]}
        )
        _destag_variable(mock_da)
    assert (
        exc_info.value.args[0] == f'Parameter datavar must be xarray.Variable, not {type(mock_da)}'
    )


def test_destag_variable_unstaggered():
    with pytest.raises(ValueError) as exc_info:
        _destag_variable(xr.Variable(('x', 'y'), np.zeros((2, 2))))
    assert (
        exc_info.value.args[0]
        == 'No dimension available to destagger. This variable does not appear to be staggered.'
    )


def test_destag_variable_missing_dim():
    with pytest.raises(ValueError):
        _destag_variable(xr.Variable(('x', 'y'), np.zeros((2, 2))), 'z_stag')


def test_destag_variable_multiple_dims():
    with pytest.raises(NotImplementedError):
        _destag_variable(xr.Variable(('x_stag', 'y_stag'), np.zeros((2, 2))))


@pytest.mark.parametrize(
    'unstag_dim_name, expected_output_dim_name',
    [
        ('z', 'z'),
        (None, 'bottom_top'),
    ],
)
def test_destag_variable_1d(unstag_dim_name, expected_output_dim_name):
    staggered = xr.Variable(
        ('bottom_top_stag',), np.arange(5), attrs={'foo': 'bar', 'stagger': 'Z'}
    )
    output = _destag_variable(staggered, unstag_dim_name=unstag_dim_name)
    # Check values
    np.testing.assert_array_almost_equal(output.values, 0.5 + np.arange(4))
    # Check dim name
    assert output.dims[0] == expected_output_dim_name
    # Check attrs
    assert output.attrs == {'foo': 'bar'}


def test_destag_variable_2d():
    staggered = xr.Variable(('x', 'y_stag'), np.arange(9).reshape(3, 3))
    expected = xr.Variable(('x', 'y'), [[0.5, 1.5], [3.5, 4.5], [6.5, 7.5]])
    xr.testing.assert_equal(_destag_variable(staggered), expected)



import xarray as xr


def _drop_attrs(attrs_dict, attrs_to_drop):
    try:
        return {k: v for k, v in attrs_dict.items() if k not in attrs_to_drop}
    except AttributeError:
        return None


def _destag_variable(datavar, stagger_dim=None, unstag_dim_name=None):
    """
    Destaggering function for a single wrf xarray.Variable

    Based off of the wrf-python destagger function
    https://github.com/NCAR/wrf-python/blob/22fb45c54f5193b849fdff0279445532c1a6c89f/src/wrf/destag.py
    Copyright 2016 University Corporation for Atmospheric Research, reused with modification
    according to the terms of the Apache License, Version 2.0

    Parameters
    ----------
    datavar : xarray.Variable
        Data variable to be destaggered
    stagger_dim : str, optional
        Name of dimension to unstagger. Defaults to guessing based on name (ends in "_stag")
    unstag_dim_name : str, option
        String to which to rename the dimension after destaggering. Example would be
        "west_east" for "west_east_stag". By default the dimenions will be renamed the text in
        front of "_stag" from the "stagger_dim" field.

    Returns
    -------
    xarray.Variable
        The destaggered variable with renamed dimension
    """
    if not isinstance(datavar, xr.Variable):
        # Implementation expects a Variable; don't want a DataArray or other type to slip through
        raise ValueError(f'Parameter datavar must be xarray.Variable, not {type(datavar)}')

    # Determine dimension to unstagger
    if stagger_dim and stagger_dim not in datavar.dims:
        # If user provided, but not actually there, error out
        raise ValueError(f'{stagger_dim} not in {datavar.dims}')
    elif stagger_dim is None:
        # If not provided, guess based on name
        stagger_dim = [x for x in datavar.dims if x.endswith('_stag')]

        if len(stagger_dim) > 1:
            raise NotImplementedError(
                'Expected a single destagger dimensions. Found multiple destagger dimensions: '
                f'{stagger_dim}'
            )
        elif len(stagger_dim) == 0:
            raise ValueError(
                'No dimension available to destagger. This variable does not appear to be '
                'staggered.'
            )

        # we need a string, not a list containing a string
        stagger_dim = stagger_dim[0]
    # Otherwise, we have a valid user provided stagger dimension

    # Destagger by mean of offset slices representing each side with respect to the stagger_dim
    stagger_dim_size = datavar.sizes[stagger_dim]
    left_or_bottom_cells = datavar.isel({stagger_dim: slice(0, stagger_dim_size - 1)})
    right_or_top_cells = datavar.isel({stagger_dim: slice(1, stagger_dim_size)})
    center_mean = (left_or_bottom_cells + right_or_top_cells) * 0.5

    # Determine new dimension name; if not given, use part of original name before "_stag"
    if unstag_dim_name is None:
        unstag_dim_name = stagger_dim.split('_stag')[0]

    # Return a Variable with renamed dimensions, updated data and attrs, and original encoding
    return xr.Variable(
        dims=tuple(str(unstag_dim_name) if dim == stagger_dim else dim for dim in center_mean.dims),
        data=center_mean.data,
        attrs=_drop_attrs(datavar.attrs, ('stagger', 'c_grid_axis_shift')),
        encoding=datavar.encoding,
        fastpath=True,
    )


def _rename_staggered_coordinate(name, stagger_dim=None, unstag_dim_name=None):
    if name == stagger_dim and unstag_dim_name is not None:
        return unstag_dim_name
    elif name[-2:].lower() in ('_u', '_v'):
        return name[:-2]
    elif name[-5:].lower() == '_stag':
        return name[:-5]
    else:
        return name



from __future__ import annotations  # noqa: F401

import xarray as xr

from .destagger import _destag_variable, _rename_staggered_coordinate
from .postprocess import (
    _assign_coord_to_dim_of_different_name,
    _calc_base_diagnostics,
    _collapse_time_dim,
    _decode_times,
    _include_projection_coordinates,
    _make_units_pint_friendly,
    _modify_attrs_to_cf,
    _rename_dims,
)


class WRFAccessor:
    """
    Common Dataset and DataArray accessor functionality.
    """

    def __init__(self, xarray_obj: xr.Dataset | xr.DataArray) -> WRFAccessor:
        self.xarray_obj = xarray_obj


@xr.register_dataarray_accessor('xwrf')
class WRFDataArrayAccessor(WRFAccessor):
    """Adds a number of WRF specific methods to xarray.DataArray objects."""

    def destagger(
        self,
        stagger_dim: str | None = None,
        unstaggered_dim_name: str | None = None,
        exclude_staggered_auxiliary_coords: bool = False,
    ) -> xr.DataArray:
        """
        Destagger a single WRF xarray.DataArray

        Parameters
        ----------
        stagger_dim : str, optional
            Name of dimension to unstagger. Defaults to guessing based on name (ends in "_stag")
        unstaggered_dim_name : str, optional
            String to which to rename the dimension after destaggering. Example would be
            "west_east" for "west_east_stag". By default the dimenions will be renamed the text in
            front of "_stag" from the "stagger_dim" field.
        exclude_staggered_auxiliary_coords : bool, optional
            If True, auxiliary coordinates (such as latitude and longitude) that originally include
            staggered dimensions are excluded from being destaggered along with this DataArray's
            data. If False, auxiliary coordinates are destaggered in the same fashion as the data
            (average of cell boundary values), which may introduce error in comparison to the
            analogous non-staggered coordinates from the original WRF data. Defaults to False.

        Returns
        -------
        xarray.DataArray
            The destaggered DataArray with renamed dimension and adjusted coordinates.

        Notes
        -----
        While destaggered coordinate variables are made available by default
        (exclude_staggered_auxiliary_coords=False), these are approximations made from the
        staggered coordinate values and may not be sufficiently accurate for all grid projections
        and/or use cases. For full accuracy, auxiliary coordinates should be re-computed from
        dimension coordinates or obtained from the original dataset.
        """
        new_variable = _destag_variable(
            self.xarray_obj.variable, stagger_dim=stagger_dim, unstag_dim_name=unstaggered_dim_name
        )

        # Need to recalculate staggered coordinates, as they don't already exist independently
        # in a DataArray context
        new_coords = {}
        for coord_name, coord_data in self.xarray_obj.coords.items():
            if set(coord_data.dims).difference(set(new_variable.dims)):
                # Has a dimension not in the destaggered output (and so still staggered)
                new_name = _rename_staggered_coordinate(
                    coord_name, stagger_dim=stagger_dim, unstag_dim_name=unstaggered_dim_name
                )
                if not exclude_staggered_auxiliary_coords or new_name in new_variable.dims:
                    # Skip if excluding and this isn't a dimension coordinate of output
                    new_coords[new_name] = _destag_variable(
                        coord_data.variable,
                        stagger_dim=stagger_dim,
                        unstag_dim_name=unstaggered_dim_name,
                    )
            else:
                new_coords[coord_name] = coord_data.variable

        return xr.DataArray(new_variable, coords=new_coords)


@xr.register_dataset_accessor('xwrf')
class WRFDatasetAccessor(WRFAccessor):
    """Adds a number of WRF specific methods to xarray.Dataset objects."""

    def postprocess(
        self,
        decode_times: bool = True,
        calculate_diagnostic_variables: bool = True,
        drop_diagnostic_variable_components: bool = True,
    ) -> xr.Dataset:
        """
        Postprocess the dataset. This method will perform the following operations:

        - Rename dimensions to match the CF conventions.
        - Rename variables to match the CF conventions.
        - Rename variable attributes to match the CF conventions.
        - Convert units to Pint-friendly units.
        - Decode times.
        - Include projection coordinates.
        - Collapse time dimension.

        Parameters
        ----------
        decode_times : bool, optional
            Decode the string-like wrfout times to xarray-friendly Pandas types. Defaults to True.
        calculate_diagnostic_variables : bool, optional
            Calculates essential diagnostic variables (potential temperature, air pressure,
            geopotential, and geopotential height) that are otherwise only present in wrfout files
            as split components or dependent upon special adjustments. Also calculates earth-relative
            wind fields, as winds by default are grid-relative. Defaults to True. If the
            underlying fields on which any of these calculated fields depends is missing, that
            calculated variable is skipped. These will be eagerly evalulated, unless your data has
            been chunked with Dask, in which case these fields will also be Dask arrays.
        drop_diagnostic_variable_components : bool, optional
            Determine whether to drop the underlying fields used to calculate the diagnostic
            variables. Defaults to True. Never drops grid-relative wind fields.

        Returns
        -------
        xarray.Dataset
            The postprocessed dataset.
        """
        ds = (
            self.xarray_obj.pipe(_modify_attrs_to_cf)
            .pipe(_make_units_pint_friendly)
            .pipe(_collapse_time_dim)
            .pipe(_assign_coord_to_dim_of_different_name)
        )
        if decode_times:
            ds = ds.pipe(_decode_times)
        if calculate_diagnostic_variables:
            ds = ds.pipe(_calc_base_diagnostics, drop=drop_diagnostic_variable_components)

        ds = ds.pipe(_include_projection_coordinates)

        return ds.pipe(_rename_dims)

    def destagger(self, staggered_to_unstaggered_dims: dict[str, str] | None = None) -> xr.Dataset:
        """
        Destagger all data variables in a WRF xarray.Dataset

        Parameters
        ----------
        staggered_to_unstaggered_dims : dict, optional
            Mapping of target staggered dimensions to corresponding unstaggered dimensions

        Returns
        -------
        xarray.Dataset
            The destaggered dataset.

        Notes
        -----
        Does not alter coordinates, only data variables. Staggered coordinates will remain on the
        dataset, but will not be associated with any data variables.
        """
        staggered_dims = (
            {dim for dim in self.xarray_obj.dims if dim.endswith('_stag')}
            if staggered_to_unstaggered_dims is None
            else set(staggered_to_unstaggered_dims)
        )
        new_data_vars = {}
        for var_name, var_data in self.xarray_obj.data_vars.items():
            if this_staggered_dims := set(var_data.dims).intersection(staggered_dims):
                # Found a staggered dim
                # TODO: should we raise an error if somehow end up with more than just one
                # staggered dim, or just pick one from the set like below?
                this_staggered_dim = this_staggered_dims.pop()
                new_data_vars[var_name] = _destag_variable(
                    var_data.variable,
                    stagger_dim=this_staggered_dim,
                    unstag_dim_name=(
                        None
                        if staggered_to_unstaggered_dims is None
                        else staggered_to_unstaggered_dims[this_staggered_dim]
                    ),
                )
            else:
                # No staggered dims
                new_data_vars[var_name] = var_data.variable

        return xr.Dataset(new_data_vars, self.xarray_obj.coords, self.xarray_obj.attrs)



import pandas as pd
import pytest
import xarray as xr

import xwrf

from . import importorskip


@pytest.fixture(scope='session')
def test_grid(request):
    return xwrf.tutorial.open_dataset(request.param)


@importorskip('cf_xarray')
@pytest.mark.parametrize(
    'name, cf_grid_mapping_name', [('lambert_conformal', 'lambert_conformal_conic')]
)
def test_postprocess(name, cf_grid_mapping_name):
    # Verify initial/raw state
    raw_ds = xwrf.tutorial.open_dataset(name)
    assert pd.api.types.is_string_dtype(raw_ds.Times.dtype)
    assert pd.api.types.is_numeric_dtype(raw_ds.Time.dtype)
    assert 'time' not in raw_ds.cf.coordinates
    assert raw_ds.cf.standard_names == {}

    # Postprocess without decoding times
    dsa = raw_ds.xwrf.postprocess(decode_times=False)
    assert pd.api.types.is_numeric_dtype(dsa.Time.dtype)

    # Postprocess
    ds = raw_ds.xwrf.postprocess()

    # Check for time coordinate handling
    assert pd.api.types.is_datetime64_dtype(ds.Time.dtype)
    assert 'time' in ds.cf.coordinates

    # Check for projection handling
    assert ds['wrf_projection'].attrs['grid_mapping_name'] == cf_grid_mapping_name

    # Check for standard name and variable handling
    standard_names = ds.cf.standard_names
    assert 'x' in standard_names['projection_x_coordinate']
    assert 'y' in standard_names['projection_y_coordinate']
    assert 'z' in standard_names['atmosphere_hybrid_sigma_pressure_coordinate']
    assert standard_names['time'] == ['Time']
    assert standard_names['humidity_mixing_ratio'] == ['Q2', 'QVAPOR']
    assert standard_names['air_temperature'] == ['T2']

    # Check for time dimension reduction
    assert ds['z'].shape == (39,)
    assert ds['z_stag'].shape == (40,)
    assert ds['XLAT'].shape == ds['XLONG'].shape == (29, 31)
    assert ds['XLAT_U'].shape == ds['XLONG_U'].shape == (29, 32)
    assert ds['XLAT_V'].shape == ds['XLONG_V'].shape == (30, 31)

    # Check for diagnostic variable calculation
    assert 'air_potential_temperature' in ds.data_vars
    assert 'air_pressure' in ds.data_vars
    assert 'geopotential' in ds.data_vars
    assert 'geopotential_height' in ds.data_vars
    assert 'T' not in ds.data_vars
    assert 'P' not in ds.data_vars
    assert 'PB' not in ds.data_vars
    assert 'PH' not in ds.data_vars
    assert 'PHB' not in ds.data_vars


@pytest.mark.parametrize('test_grid', ['lambert_conformal', 'mercator'], indirect=True)
def test_dataarray_destagger(test_grid):
    data = test_grid['U']
    destaggered = data.xwrf.destagger()

    # Check shape reduction and dim name adjustment
    assert destaggered.sizes['west_east'] == data.sizes['west_east_stag'] - 1

    # Check coordinate reduction
    xr.testing.assert_allclose(destaggered['XLAT'], test_grid['XLAT'])
    xr.testing.assert_allclose(destaggered['XLONG'], test_grid['XLONG'])

    # Check attributes are preserved
    assert set(destaggered.attrs.keys()) == set(data.attrs.keys()) - {
        'stagger',
    }


@pytest.mark.parametrize('test_grid', ['lambert_conformal', 'mercator'], indirect=True)
def test_dataarray_destagger_with_exclude(test_grid):
    data = test_grid['V']
    destaggered = data.xwrf.destagger(exclude_staggered_auxiliary_coords=True)

    # Check shape reduction and dim name adjustment
    assert destaggered.sizes['south_north'] == data.sizes['south_north_stag'] - 1

    # Verify that no XLAT/XLONG are present in output, but XTIME (which is auxiliary, but not
    # staggered) remains
    assert not any(coord_name.startswith('XL') for coord_name in destaggered.coords)
    assert 'XTIME' in destaggered.coords


@pytest.mark.parametrize('test_grid', ['lambert_conformal', 'mercator'], indirect=True)
def test_dataarray_destagger_with_postprocess(test_grid):
    postprocessed = test_grid.xwrf.postprocess()
    data = postprocessed['U']
    destaggered = data.xwrf.destagger(exclude_staggered_auxiliary_coords=True)

    # Check staggered to unstaggered dimension coordinate handling
    xr.testing.assert_allclose(destaggered['x'], postprocessed['x'])

    # Check attributes are preserved, other than those made invalid by destaggering
    assert set(destaggered.attrs.keys()) == set(data.attrs.keys()) - {
        'stagger',
    }
    assert 'c_grid_axis_shift' not in destaggered['x'].attrs


@pytest.mark.parametrize('test_grid', ['lambert_conformal', 'mercator'], indirect=True)
def test_dataset_destagger(test_grid):
    destaggered = (
        test_grid.isel(Time=slice(0, 2))
        .xwrf.postprocess(calculate_diagnostic_variables=False)
        .xwrf.destagger()
    )

    # Check elimination of staggered dims and "stagger" attr
    for varname in destaggered.data_vars:
        assert not {'x_stag', 'y_stag', 'z_stag'}.intersection(set(destaggered[varname].dims))
        assert (
            'stagger' not in destaggered[varname].attrs
            or destaggered[varname].attrs['stagger'] == ''
        )

    # Check preservation of variable attrs
    for varname in set(test_grid.data_vars).intersection(set(destaggered.data_vars)):
        # because of xwrf.postprocess, the destaggered attrs will include more information
        assert set(test_grid[varname].attrs.keys()) - {'stagger', 'units'} <= set(
            destaggered[varname].attrs.keys()
        )

    # Check that attrs are preserved
    assert destaggered.attrs == test_grid.attrs
